//messages.java Q1 2004
//Features:
// maintains list of messages
// messages can be read or unread
// deletes messages after specified time

import java.io.*;
import java.util.*;

import java.util.Timer;
import java.util.TimerTask;

public class messages
{
	static final int version = 2;

	//Modules
	static loggers   lm;
	static config    cfg;
	static data      pd;
	static tcpserver net;
	static dbio      io;

	static Hashtable messages;

	static java.text.SimpleDateFormat messageformat;
	
	public messages(loggers lm, config cfg, data pd, tcpserver net, dbio io)
	{
		this.lm  = lm;
		this.cfg = cfg;
		this.pd  = pd;
		this.net = net;
		this.io  = io;
		
		messages = new Hashtable();
		
		messageformat = new java.text.SimpleDateFormat(
			"MMM dd HH:mm",
			new java.util.Locale("en","US").getDefault()
			);

		//Set up a timer callback for persist
		Timer persistTimer = new Timer(true); //daemon
		int seconds = cfg.GetInt("Persist", "SyncSeconds", 300) * 1000;
		persistTimer.schedule(new PersistTimer(this), seconds, seconds);
	}

//// Misc util ////

	//Seconds
	protected long getTime()
	{
		return new java.util.Date().getTime() / 1000;
	}

//// Core ////

	public void storeMessage(String port, String pid, String to,
		String message)
	{
		if (to == null) return;
		String from = pd.getNameByPid(port, pid);
		String hash = to + from;
		Message m = (Message)messages.get(hash.toLowerCase());
		if (m == null)
		{
			m = new Message();
			messages.put(hash.toLowerCase(), m);
			net.SendMessage(port, pid, "Message stored");
		}
		else
			net.SendMessage(port, pid, "Previous message replaced");
		
		m.to      = to;
		m.from    = from;
		m.message = message;
		m.time    = getTime();
		
		//If the target is online, tell them now
		Enumeration ports = pd.getPorts(to);
		Enumeration pids  = pd.getPids(to);
		if (ports == null || pids == null)
		{
			//Offline
			return;
		}
			
		while (ports.hasMoreElements() &&
			pids.hasMoreElements())
		{
			int count = getNewMessageCount(to);
			net.SendMessage((String)ports.nextElement(),
				(String)pids.nextElement(), "You have " + count
				+ " new message. Type ?messages to read");
		}
	}

	public int getNewMessageCount(String name)
	{
		Message m;
		int count = 0;

		Enumeration e = messages.elements();
		while(e.hasMoreElements())
		{
			m = (Message)e.nextElement();
			if (name.equalsIgnoreCase(m.to))
			{
				if (m.read == 0)
					count++;
			}
		}
		
		return count;
	}

	public int getMessageCount(String name)
	{
		Message m;
		int count = 0;

		Enumeration e = messages.elements();
		while(e.hasMoreElements())
		{
			m = (Message)e.nextElement();
			if (name.equalsIgnoreCase(m.to))
				count++;
		}
		
		return count;
	}

	public void checkMessages(String port, String pid)
	{
		Message m;
		String name = pd.getNameByPid(port, pid);
		boolean bNoMessages = true;
		String time;

		Enumeration e = messages.elements();
		while(e.hasMoreElements())
		{
			m = (Message)e.nextElement();
			if (name.equalsIgnoreCase(m.to))
			{
				time = messageformat.format(
					new java.util.Date(m.time*1000)
					);
			
				net.SendMessage(port, pid, "[" + time + "] " +
					m.from + ": " + m.message);
				
				bNoMessages = false;
				if (m.read == 0)
					m.read = getTime();
			}
		}
	
		if (bNoMessages)
			net.SendMessage(port, pid, "No messages");
	}

	public boolean hasExpired(Message m)
	{
		long now = getTime();
		
		String str = cfg.GetStr("Message", "Expire");
		if (str == null || str.length() < 1)
			str = "168";
		long expire = Long.parseLong( str ) * 3600;
		
		str = cfg.GetStr("Message", "ExpireRead");
		if (str == null || str.length() < 1)
			str = "48";
		long expireread = Long.parseLong( str ) * 3600;

		if (m.read > 0)
		{
			if (now > m.read + expireread)
				return true;
			return false;
		}
		
		if (now > m.time + expire)
			return true;
	
		return false;
	}

//// File IO ////

	public void load()
	{
		loadMessages("message.dat");
	}

	public void flush()
	{
		flushMessages("message.dat");
	}

	//message.dat
	public void loadMessages(String fname)
	{
		//lm.LogM("D", "messages", "Reading " + fname);
		BufferedReader in = io.fopenr(fname);
		if (in == null) return;

		Message m = null;
		String to = "*Unknown*", hash;

		String[] part;
		String key, value;
		while(true)
		{
			part = io.readLine(in);
			if (part == null) break;
			key   = part[0];
			value = part[1];
			
			if (key.equalsIgnoreCase("TO"))
			{
				to = value;
			}
			else if (key.equalsIgnoreCase("FROM"))
			{
				hash = to + value;
				m = (Message)messages.get(hash.toLowerCase());
				if (m == null)
				{
					m = new Message();
					messages.put(hash.toLowerCase(), m);
				}
				m.to   = to;
				m.from = value;
			}
			else if (key.equalsIgnoreCase("MSG"))
			{
				m.message = value;
			}
			else if (key.equalsIgnoreCase("TIME"))
			{
				m.time = Long.parseLong( value );
			}
			else if (key.equalsIgnoreCase("READ"))
			{
				m.read = Long.parseLong( value );
			}
		}
		
		io.close(in);
		lm.LogM("I", "messages", "message.dat  " + messages.size());
	}

	//message.dat
	public void flushMessages(String fname)
	{
		PrintWriter out = io.fopenw(fname);
		if (out == null) return;
		Message m;

		Enumeration e = messages.elements();
		while(e.hasMoreElements())
		{
			m = (Message)e.nextElement();
			out.println("TO:"  + m.to);
			out.println("FROM:" + m.from);
			out.println("MSG:" + m.message);
			out.println("TIME:" + m.time);
			if (m.read > 0)
				out.println("READ:" + m.read);
		}

		out.close();
	}

//// Persist ////

	class PersistTimer extends TimerTask
	{
		messages mesg;
	
		public PersistTimer(messages mesg)
		{
			this.mesg = mesg;
		}
	
		public void run()
		{
			//Remove old messages
			Message m;
			String hash;

			Enumeration e = messages.elements();
			while(e.hasMoreElements())
			{
				m = (Message)e.nextElement();
				if (hasExpired(m))
				{
					hash = m.to + m.from;
					messages.remove(hash.toLowerCase());
				}
			}
		
			lm.LogM("D", "messages", "Saving persistent data");
			mesg.flush();
		}
	}

//// Containers ////

	class Message
	{
		static final int version = 1;

		public String to;
		public String from;
		public long   time;
		public long   read;    //Time when user read the message
		public String message;
		
		public Message()
		{
			to      = null;
			from    = null;
			time    = 0; //getTime();
			read    = 0;
			message = null;
		}
	}
}
